/*
 * Copyright 2020 Google LLC
 *
 * Use of this source code is governed by a BSD-style
 * license that can be found in the LICENSE file or at
 * https://developers.google.com/open-source/licenses/bsd
 */

#ifndef REFTABLE_CONSTANTS_H
#define REFTABLE_CONSTANTS_H

#define REFTABLE_BLOCK_TYPE_LOG 'g'
#define REFTABLE_BLOCK_TYPE_INDEX 'i'
#define REFTABLE_BLOCK_TYPE_REF 'r'
#define REFTABLE_BLOCK_TYPE_OBJ 'o'
#define REFTABLE_BLOCK_TYPE_ANY 0

#endif /* REFTABLE_CONSTANTS_H */
