/*
 * Copyright 2020 Google LLC
 *
 * Use of this source code is governed by a BSD-style
 * license that can be found in the LICENSE file or at
 * https://developers.google.com/open-source/licenses/bsd
 */

#ifndef CONSTANTS_H
#define CONSTANTS_H

#include "reftable-constants.h"

#define MAX_RESTARTS ((1 << 16) - 1)
#define DEFAULT_BLOCK_SIZE 4096
#define DEFAULT_GEOMETRIC_FACTOR 2

#endif
